#!/usr/bin/awk -f 
# <License_blurb>
#
# This code is released under the X11 License. 
# 
# Copyright (C) 2006 Christoph Mitterhofer <c.mitterhofer@gmail.com>
# 
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# </License blurp>
# 
# HISTORY: 
#    cmi 2006-12-19: 
#      * Commented the code.
#      * Simplified printattr()
#      * Found and fixed a bug in the output, that made this
#        incompatible with xml (search for the comment containing
#        "2xml"). 
#      * Commented the code. Did i already say, that i commented the code? 
#      * Inserted in a license notice. 
#      
#    cmi earlier in 2006: 
#      wrote the first version for m.gaugusch's fou4s. 
# 
# This is a xml2-alike in awk. It's not the same. Feeding the output to 
# 2xml may lead to unexpected behaviour. 
# 

# if we define '<' as Record separator, then each "record" starts either 
# with an opening or a closing tag. That's the basic idea. 
BEGIN{
  # Record separator 
  RS="<";
  # Output "path" separator
  XPATH="/";
  depth = 0
  # XLIST is used to store the current "path" 
  XLIST[0]=""
}
#
#
# function tagline
# returns:
#   The current path. This is composed from all path elements stored in 
#   XLIST. 
function tagline() { 
 pstr = ""; 
 for (i=1; i<=depth;i++) {pstr=pstr "/" XLIST[i];}
 return pstr; 
}
#
#
# function printattr
#   prints attributes in XML2-notation. 
# args: 
#   prefix: The value returned by function tagline()
#   arr: Array containing '$attname="$attval"' pairs
# local variables: 
#   i: place-holder variable
function printattr(prefix, arr,      i) { 
  for (i in arr) { 
    l = arr[i]; 
    gsub(/=\"/, "=", l ) 
    gsub(/\"$/, "", l ) 
    print prefix l 
  }
}
#
#
# main
# 
# The first token in $0 is always a tag name (either "$tag" or "/$tag>") 
{ 
  # Remove trailing whitespace
  gsub(/>[\n ]*$/, ">")
  
  if ($1=="?xml" ) { 
    # The <?xml/> header is silently ignored if it exists. 
  } else if ( $1 ~ /!--/ ) { 
    # xml-comments are printed, verbatim. 
    print $0 ; 
  } else if ($1~/^\//) { 
    # The first character is a '/', so this is an end-tag: </tag>
    # Extract the tagname, strip '/', '>' and trailing whitespace
    tag=$1
    gsub(/^\//, "", tag)
    gsub(/> *$/, "", tag)
    # Check, if the XML is well-formed.  The extracted tag should 
    # match the value stored in XLIST. 
    otag = XLIST[depth]
    if (tag != otag) { 
      # XML is not well-formed. die! 
      print "Expected </" otag ">, got '" tag "'"
      exit(1)
    }
    # go one up in the hierarchy (XLIST) 
    depth--; 
  } else if ($1~/^[a-zA-Z0-9:]/) { 
    # $0 starts with a "word" character, so it MUST be a start tag. 
    emptytag = 0  # 0: no, 1: yes, 2: yes, but following a twin
    tag = $1
    # Strip unneded characters from tag... or so. 
    gsub(/\/*>.*/, "", tag)
    # identify empty (<tag/>) tags. 
    if ($0~/\/>$/) {
      emptytag = 1;
    }
    # store tag in XLIST
    depth++; 
    #
    # There is a problem here. If we have two identical empty tags 
    # following each other, the output is ambiguous. 2xml (the counter-
    # part of xml2) merges these two emtpy tags into one. 
    #  The solution for xml2 is to handle all but the first empty tag
    # like a non-empty-tag, i.e. output the header line.  We can do that, too.
    if (XLIST[depth] == tag && emptytag == 1) { 
      emptytag = 2;
    }
    ####
    XLIST[depth] = tag
    # If a tag is not empty, we need some kind of "start"-marker. 
    # xml2 writes an empty path. So do we. 
    if (emptytag == 0 || emptytag == 2) { 
      print tagline();
    }
    # processing of attributes and content.  
    content = $0
    $1 = ""
    attribs = $0
    # Strip content from attributes. 
    gsub(/[a-z:0-9]*\/*>.*/, "", attribs)
    # strip leading whitespace from attributes
    gsub(/^ */, "", attribs)

    # strip attributes from content. 
    gsub(/^[^>]*>/, "", content)

    # Handle attributes. Attributes in well-formed xml are of the form
    # $attname="$attval", so '" ' should be a good separator. 
    if (attribs != "") {
      # ... thus we split attribs on '" ' and print the result with
      # printattr()
      fields = split(attribs, att, "\" ")
      prefix = tagline() "/@"; 
      printattr(prefix, att); 
    }
    
    # Handle content. 
    if (content != "") {
      # Escape =
      gsub("=", "", content); 
      # Replace Newlines by Ctrl-E
      gsub(" *\n", "", content); 
      # Remove spurious ^M 
      gsub("
", "", content); 
      # and print the resulting content-line. 
      print tagline() "=" content;
    }
    # When in an empty tag, depth is now wrong. Correct this. 
    if (emptytag == 1 || emptytag == 2) { depth--; }
  } else if ($0 == "") { ; 
    # This happens sometimes. I'm ignoring this. 
  } else { 
    # The classical "this should never happen"-error. 
    # If you get this, then either you found a bug in this 130+ lines of 
    # code, or your input was garbage. 
    print "Garbage: " $0; 
    exit(1)
  }
}
